/* linuxSampler.c */
/* linuxSampler.c,v 1.1.1.1 2005/11/03 19:58:57 xusf Exp */
/***********************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of
* California, as Operator of Los Alamos National Laboratory, and
* Berliner Elektronenspeicherring-Gesellschaft m.b.H. (BESSY).
* asynDriver is distributed subject to a Software License Agreement
* found in file LICENSE that is included with this distribution.
***********************************************************************/
/*Linux dependent file
 * Author: Shifu Xu
 * Date:   03NOV2005
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <iocsh.h>
#include <epicsEvent.h>
#include <cantProceed.h>
#include <epicsExport.h>
#include <devLib.h>

#include "sampler_linux.h"
#include "sampler.h"
#define epicsExportSharedSymbols
#define SAMPLER_DEVICE "/dev/sampler_device"
static void samplerSignalAction (int sig, siginfo_t * siginfo, void *extra);

volatile int useUserLevelInterrupt = 1;
epicsExportAddress (int, useUserLevelInterrupt);
typedef struct samplerOncePvt
{
    int sampler_handle;
    double clockRateMhz;
    epicsEventId interruptEvent;
    struct timeval tsUserInterrupt;
} samplerOncePvt;
static samplerOncePvt *psamplerOncePvt;

static unsigned long
diffTime (const struct timeval *end, const struct timeval *start)
{
    long sec;
    long usec;

    sec = end->tv_sec - start->tv_sec;
    usec = end->tv_usec - start->tv_usec;

    if (usec < 0)
      {
	  sec -= 1;
	  usec += 1000000;
      }
    return sec * 1000000 + usec;

}

static void samplerIntService (void *pvt)
{
    gettimeofday (&psamplerOncePvt->tsUserInterrupt, NULL);
    epicsEventSignal (psamplerOncePvt->interruptEvent);
}

static int getSample (void *pvt, delayMeasure * pdelayMeasure)
{
    struct timeval tsContextSwitch;
    samplerOncePvt *psamplerOncePvt = (samplerOncePvt *) pvt;
    struct ts_kernel tsParam;

    ioctl (psamplerOncePvt->sampler_handle,
	   SAMPLER_IOCTL_GENERATE_VME_INTERRUPT, 0);
    gettimeofday (&tsContextSwitch, NULL);
    if (useUserLevelInterrupt)
	epicsEventMustWait (psamplerOncePvt->interruptEvent);
    ioctl (psamplerOncePvt->sampler_handle, SAMPLER_IOCTL_GETTS, &tsParam);

    /* calculate */
    pdelayMeasure->interruptLatency =
	diffTime (&tsParam.top_half, &tsParam.interrupt_begin);
    pdelayMeasure->contextSwitch =
	diffTime (&tsContextSwitch, &tsParam.context_begin);
    pdelayMeasure->topToBottom =
	diffTime (&tsParam.bottom_half, &tsParam.top_half);
    if (useUserLevelInterrupt)
      {
	  pdelayMeasure->bottomToUserInterrupt =
	      diffTime (&psamplerOncePvt->tsUserInterrupt,
			&tsParam.bottom_half);
	  pdelayMeasure->totalResponse =
	      diffTime (&psamplerOncePvt->tsUserInterrupt,
			&tsParam.interrupt_begin);
      }
    else
      {
	  pdelayMeasure->bottomToUserInterrupt = 0;
	  pdelayMeasure->totalResponse =
	      diffTime (&tsContextSwitch, &tsParam.interrupt_begin);
      }
    return 0;
}

static void samplerSignalAction (int sig, siginfo_t * siginfo, void *extra)
{
    samplerIntService (psamplerOncePvt);
}

void
platformSamplerInit (const char *name, int vector, int level,
		     double clockRateMhz)
{
    data_from_user *psamplerToKernel;
    struct sigaction *paction;

    psamplerToKernel = callocMustSucceed (1, sizeof (data_from_user), "name");
    psamplerToKernel->level = level;
    psamplerToKernel->vector = vector;
    psamplerToKernel->signalEvent.sigev_signo = SIGIO;
    psamplerToKernel->signalEvent.sigev_notify = SIGEV_SIGNAL;
    psamplerToKernel->signalEvent.sigev_value.sival_int = 0;

    psamplerOncePvt = callocMustSucceed (1, sizeof (samplerOncePvt), "name");
    psamplerOncePvt->clockRateMhz = clockRateMhz;
    psamplerOncePvt->interruptEvent = epicsEventCreate (epicsEventEmpty);
    psamplerOncePvt->sampler_handle = open (SAMPLER_DEVICE, O_RDWR);
    if (psamplerOncePvt->sampler_handle == -1)
      {
	  printf ("can't open SAMPLER_DEVICE !,");
	  printf ("errno is %s,%d \n", strerror (errno), errno);
	  return;
      }
    if (ioctl
	(psamplerOncePvt->sampler_handle, SAMPLER_IOCTL_ATTACH,
	 psamplerToKernel))
      {
	  printf ("sampler ioctl error !");
	  return;
      }

    psamplerManager->registerSampler (name, getSample, psamplerOncePvt);

    if (useUserLevelInterrupt)
	ioctl (psamplerOncePvt->sampler_handle,
	       SAMPLER_IOCTL_USEUSERLEVELINTERRUPT, 1);
    else
	ioctl (psamplerOncePvt->sampler_handle,
	       SAMPLER_IOCTL_USEUSERLEVELINTERRUPT, 0);

    if (useUserLevelInterrupt)
      {
	  paction = (struct sigaction *) malloc (sizeof (struct sigaction));
	  sigemptyset (&paction->sa_mask);
	  paction->sa_sigaction = samplerSignalAction;
	  paction->sa_flags = SA_SIGINFO;
	  sigaction (SIGIO, paction, NULL);
      }
}

void platformSamplerQuit (void *pvt)
{
    samplerOncePvt *psamplerOncePvt = (samplerOncePvt *) pvt;

    psamplerManager->removeSampler (pvt);
    free (psamplerOncePvt);
}

static const iocshArg initArg0 = { "platformSamplerInit", iocshArgString };
static const iocshArg initArg1 = { "interrupt vector", iocshArgInt };
static const iocshArg initArg2 = { "interrupt level", iocshArgInt };
static const iocshArg initArg3 = { "clock rate in Mhz", iocshArgDouble };

/* may need more arguments */
static const iocshArg *const initArgs[4] =
    { &initArg0, &initArg1, &initArg2, &initArg3 };
static void initCallFunc (const iocshArgBuf * args)
{
    platformSamplerInit (args[0].sval, args[1].ival, args[2].ival,
			 args[3].dval);
}

static const iocshFuncDef initFuncDef = { "platformSamplerInit", 4, initArgs };

static void platformSamplerRegister (void)
{
    static int firstTime = 1;
    if (!firstTime)
	return;
    firstTime = 0;
    iocshRegister (&initFuncDef, initCallFunc);
}

epicsExportRegistrar (platformSamplerRegister);
